
/* Generated data (by glib-mkenums) */

#ifndef META_ENUM_TYPES_H
#define META_ENUM_TYPES_H

#include <glib-object.h>

G_BEGIN_DECLS

GType meta_theme_error_get_type (void) G_GNUC_CONST;
#define META_TYPE_THEME_ERROR (meta_theme_error_get_type())

GType meta_theme_type_get_type (void) G_GNUC_CONST;
#define META_TYPE_THEME_TYPE (meta_theme_type_get_type())

G_END_DECLS

#endif

/* Generated data ends here */

